__author__ = 'luka'
import tempfile
from nose.tools import *

from sbdk.error import SbdkError
from sbdk.project import destination, sanitize_project_name


def test_destination():
    cwd = tempfile.mkdtemp()
    empty = destination(cwd, '')
    eq_(cwd, empty, "Destination should stay unchanged")


def test_sanitize_valid():
    for valid in 'valid', 'val1d', 'va_lid':
        result = sanitize_project_name(valid)
        eq_(valid, result, "Valid name should stay unchanged")


def test_sanitize_minus():
    minus = sanitize_project_name('has-minus')
    eq_(minus, 'has_minus', "Minus should change to underscore")


@raises(SbdkError)
def test_sanitize_keyword():
    keyword = sanitize_project_name('while')


@raises(SbdkError)
def test_sanitize_pattern():
    illegal = sanitize_project_name('1asd')



from sbgsdk.wrapper import Wrapper, _get_method_requirements, Resources


error = NotImplementedError('Methods split, work and merge must be overridden.')


class ScatterGatherWrapper(Wrapper):
    def _entry_point(self):
        return self.job('_split', requirements=self.get_split_requirements(), name='split')

    def _split(self):
        args_list = self.split()
        jobs = [self.job('_work', requirements=self.get_work_requirements(args),
                         name='work_%s' % ndx, args={'job': args})
                for ndx, args in enumerate(args_list)]
        return self.job('_merge', name='merge', requirements=self.get_merge_requirements(),
                        args={'job_results': jobs})

    def _work(self, job):
        return self.work(job)

    def _merge(self, job_results):
        return self.merge(job_results)

    def split(self):
        raise error

    def work(self, job):
        raise error

    def merge(self, job_results):
        raise error

    def get_split_requirements(self):
        return _get_method_requirements(self, 'split') or Resources()

    def get_work_requirements(self, job):
        return _get_method_requirements(self, 'work') or Resources()

    def get_merge_requirements(self):
        return _get_method_requirements(self, 'merge') or Resources()




def find_image(client, repo, tag='latest'):
    images = client.images(repo)
    images = filter(lambda x: (repo + ':' + tag) in x['RepoTags'], images)
    return (images or [None])[0]


def parse_repository_tag(repo):
    column_index = repo.rfind(':')
    if column_index < 0:
        return repo, ""
    tag = repo[column_index+1:]
    slash_index = tag.find('/')
    if slash_index < 0:
        return repo[:column_index], tag

    return repo, ""



__author__ = 'luka'
import os
import shutil
import docker.auth

from sbdk.project import Project
from sbdk.docker import container
from sbdk.error import SbdkError
from sbdk.commands import init, push, test, sh, run, show_settings, schema, cf, fix_uid


BASE_CTX = {'docker_client_class': 'sbdk.tests.mock_client.MockClient'}


################################################
##
## (base_context, +id, +dir, +id+dir)
##  x (inside project dir, outside project dir)
##  x commands
##
################################################

NOTHING, IMAGE, DIR, STATE = 0, 1, 2, 4

ARGS = os.path.join(os.path.dirname(__file__), 'empty-args.json')


def run_adun(p):
    old = p.context['docker_client_class']
    try:
        p.context['docker_client_class'] = 'sbdk.tests.mock_client.OutJson'
        run.run(p, 'wrapper', ARGS, os.path.join(os.path.dirname(__file__), '__out__.json'), False)
        run.run(p, 'wrapper', ARGS, os.path.join(os.path.dirname(__file__), '__out__.json'), True)
    finally:
        p.context['docker_client_class'] = old


def run_push(p):
    if p.project_data:
        os.chdir(p.project_dir)
        with open('.sbdk/schema.json', 'w') as fp:
            fp.write('{}')
    push.push(p, 'message')


CMDS = {'run':      {'cmd': run_adun,
                     'depend': IMAGE | DIR},
        'fix-uid':  {'cmd': lambda p: fix_uid.fix_uid(p, '1000'),
                     'depend': IMAGE},
        'push':     {'cmd': run_push,
                     'depend': IMAGE | DIR | STATE},
        'schema':   {'cmd': schema.schema,
                     'depend': IMAGE},
        'show-settings': {'cmd': show_settings.show,
                          'depend': NOTHING},
        'test':     {'cmd': lambda p: test.test(p, '.'),
                     'depend': IMAGE | DIR}
}


def test_commands():
    container.push_json = lambda url, data, session_id: None
    docker.auth.auth.load_config = lambda: {'Configs': {p.docker_registry_url: {'Username': 'test', 'Password': 'test'}}}
    container.load_config = lambda: {'Configs': {p.docker_registry_url: {'Username': 'test', 'Password': 'test'}}}

    # test in project
    ctx = {'cwd': '/tmp/sbdk-test'}
    ctx.update(BASE_CTX)
    p = Project(ctx)
    try:
        p.initialize('prj_name')
        os.chdir(p.project_dir)

        for cmd in CMDS.itervalues():
            cmd['cmd'](p)
    finally:
        d = '/tmp/sbdk-test/prj_name'
        if os.path.isdir(d):
            shutil.rmtree(d)

    for conf in IMAGE, DIR, IMAGE | DIR:
        ctx = {'cwd': '/tmp/sbdk-test'}
        ctx.update(BASE_CTX)
        if conf & IMAGE:
            ctx['image_id'] = '1234567890ab'
        if conf & DIR:
            ctx['project_dir'] = 'prj'
            os.makedirs('/tmp/sbdk-test/prj')
            os.chdir('/tmp/sbdk-test')

        p = Project(ctx)

        try:
            for name, cmd in CMDS.iteritems():
                satisfied = (conf & cmd['depend']) == cmd['depend']
                try:
                    cmd['cmd'](p)
                    assert satisfied, "The command {} should fail as it's conditions are not satisfied".format(name)
                except SbdkError, e:
                    if satisfied:
                        raise
                    else:
                        print "Ignoring expected message {} for command {}".format(e.message, name)

        finally:
            if p.project_dir and os.path.isdir(p.project_dir):
                shutil.rmtree(p.project_dir)



import logging
from sbgsdk.exceptions import ValidationException, JobException, ProtocolException

from sbgsdk.job import BaseJob, Job, Outputs
from sbgsdk.util import import_name
from sbgsdk.protocol import to_json


class Executor(object):
    def execute(self, job):
        pass


class Tassadar(Executor):
    def __init__(self):
        self.results = {}
        self.job_executors = {
            Job: self.exec_wrapper_job,
        }

    def execute(self, job):
        logging.debug('Job started: %s' % to_json(job))
        if job.job_id in self.results:
            return self.results[job.job_id]
        for key, val in job.args.iteritems():
            job.resolved_args[key] = self.resolve(val)
        if job.__class__ not in self.job_executors and not callable(job):
            raise NotImplementedError('Tassadar unable to run job of type %s.' % job.__class__.__name__)
        job.status = BaseJob.RUNNING
        result = job() if callable(job) else self.job_executors[job.__class__](job)
        self.results[job.job_id] = result
        job.status = BaseJob.DONE
        if isinstance(result, BaseJob):
            return self.execute(result)
        logging.debug('Job result: %s' % to_json(result))
        return result

    def exec_wrapper_job(self, job):
        cls = import_name(job.wrapper_id)
        try:
            wrp = cls(inputs=job.resolved_args.pop('$inputs', {}),
                      params=job.resolved_args.pop('$params', {}),
                      context=job.context,
                      resources=job.resources)
            result = wrp(job.resolved_args.pop('$method', None), job.resolved_args)
        except (ValidationException, ProtocolException):
            raise
        except Exception as e:
            msg = get_exception_message(e)
            logging.exception('Job failed: %s', msg)
            raise JobException(msg)
        return result if result is not None else Outputs(wrp.outputs.__json__())

    def resolve(self, val):
        if isinstance(val, BaseJob):
            return self.execute(val)
        if isinstance(val, (list, tuple)):
            return [self.resolve(item) for item in val]
        if isinstance(val, dict):
            return {k: self.resolve(v) for k, v in val.iteritems()}
        return val


def get_exception_message(exception):
    return exception.message or str(exception)


from .validators import validate_type, validate_range, validate_enum, validate_pattern, validate_struct
from .base_attr import BaseAttr, default_attr_name


class BaseParam(BaseAttr):
    def __init__(self, name=None, description='', required=False, default=None, category=None,
                 list=False, show_if=None, **extra):
        BaseAttr.__init__(self, name, description, required, list, **extra)
        default = [] if list and default is None else default
        self.default = default() if callable(default) else default
        self.category = category
        self.condition = show_if

    def __call__(self, attr_name, data=None):
        BaseAttr.__call__(self, attr_name, data)
        data = self.default if data is None else data
        if self.list and not isinstance(data, (list, tuple)):
            data = [] if data is None else [data]
        return data


class IntAttr(BaseParam):
    def __init__(self, name=None, description='', required=False, default=None, category=None,
                 list=False, min=None, max=None, step=None, **extra):
        BaseParam.__init__(self, name, description, required, default, category, list, **extra)
        self.type = 'integer'
        self.min = min
        self.max = max
        self.step = step

    def validate(self, instance):
        BaseParam.validate(self, instance)
        validate_type(instance, (int, long), list=self.list)
        validate_range(instance, self.min, self.max, self.step, list=self.list)


class RealAttr(BaseParam):
    def __init__(self, name=None, description='', required=False, default=None, category=None,
                 list=False, min=None, max=None, step=None, **extra):
        BaseParam.__init__(self, name, description, required, default, category, list, **extra)
        self.type = 'float'
        self.min = min
        self.max = max
        self.step = step

    def validate(self, instance):
        BaseParam.validate(self, instance)
        validate_type(instance, (float, int, long), list=self.list)
        validate_range(instance, self.min, self.max, self.step, list=self.list)


class StringAttr(BaseParam):
    def __init__(self, name=None, description='', required=False, default=None, category=None,
                 list=False, pattern=None, **extra):
        BaseParam.__init__(self, name, description, required, default, category, list, **extra)
        self.type = 'string'
        self.pattern = pattern

    def validate(self, instance):
        BaseParam.validate(self, instance)
        validate_type(instance, (basestring,), list=self.list)
        validate_pattern(instance, self.pattern, list=self.list)


class BoolAttr(BaseParam):
    def __init__(self, name=None, description='', required=False, default=None, category=None,
                 list=False, **extra):
        BaseParam.__init__(self, name, description, required, default, category, list, **extra)
        self.type = 'boolean'

    def validate(self, instance):
        BaseParam.validate(self, instance)
        validate_type(instance, (bool,), list=self.list)


class EnumAttr(BaseParam):
    def __init__(self,  values, name=None, description='', required=False, default=None, category=None,
                 list=False, **extra):
        BaseParam.__init__(self, name, description, required, default, category, list, **extra)
        self.type = 'enum'
        self.values = [self._expand_value(v) for v in values]

    def _expand_value(self, val):
        error = ValueError('Please specify a 3-tuple with id, name and description.')
        d = default_attr_name
        if not isinstance(val, (tuple, list)):
            return val, d(val), d(val)
        l = len(val)
        if not l:
            raise error
        elif l == 1:
            return val[0], d(val[0]), d(val[0])
        elif l == 2:
            if not isinstance(val[1], basestring):
                raise error
            return val[0], val[1], d(val[0])
        elif l == 3:
            if not isinstance(val[1], basestring) or not isinstance(val[2], basestring):
                raise error
            return val
        else:
            raise error

    def validate(self, instance):
        BaseParam.validate(self, instance)
        validate_enum(instance, [v[0] for v in self.values], list=self.list)


class StructAttr(BaseParam):
    def __init__(self, schema, item_label=None, name=None, description='', required=False, default=None, category=None,
                 list=False, **extra):
        BaseParam.__init__(self, name, description, required, default, category, list, **extra)
        self.schema = schema
        self.item_label = item_label or schema.__name__
        self.type = 'struct'

    def validate(self, instance):
        BaseParam.validate(self, instance)
        validate_type(instance, (self.schema,), list=self.list)
        validate_struct(instance, list=self.list)

    def __call__(self, attr_name, data=None):
        to_dict = lambda o: o.__json__() if hasattr(o, '__json__') else o or {}
        BaseParam.__call__(self, attr_name, data)
        if not self.list:
            return self.schema(**to_dict(data))
        if not data:
            return []
        return [self.schema(**to_dict(item)) for item in data]


from copy import deepcopy
import signal


def wrap_in_list(val, *args):
    """
    >>> wrap_in_list(1, 2)
    [1, 2]
    >>> wrap_in_list([1, 2], 3, 4)
    [1, 2, 3, 4]
    >>> wrap_in_list([1, 2], [3, 4])
    [1, 2, [3, 4]]
    """
    wrapped = val if isinstance(val, list) else [val]
    return wrapped + list(args)


def import_name(name):
    name = str(name)
    if '.' not in name:
        return __import__(name)
    chunks = name.split('.')
    var_name = chunks[-1]
    module_name = '.'.join(chunks[:-1])
    fromlist = chunks[:-2] if len(chunks) > 2 else []
    module = __import__(module_name, fromlist=fromlist)
    if not hasattr(module, var_name):
        raise ImportError('%s not found in %s' % (var_name, module_name))
    return getattr(module, var_name)


class DotAccessDict(dict):
    def __getattr__(self, item):
        return self.get(item, None)

    def __setattr__(self, key, value):
        self[key] = value

    def __copy__(self):
        return self.__class__(self)

    def __deepcopy__(self, memo):
        return self.__class__(deepcopy(dict(self), memo))


def intersect_dicts(d1, d2):
    """
    >>> intersect_dicts({'a': 1, 'b': 2}, {'a': 1, 'b': 3})
    {'a': 1}
    >>> intersect_dicts({'a': 1, 'b': 2}, {'a': 0})
    {}
    """
    return {k: v for k, v in d1.iteritems() if v == d2.get(k)}


class SignalContextProcessor(object):
    def __init__(self, handler, *signals):
        self.handler = handler
        self.signals = signals
        self.old_handlers = {}

    def __enter__(self):
        self.old_handlers = {sig: signal.signal(sig, self.handler) for sig in self.signals}

    def __exit__(self, type, value, traceback):
        for sig in self.signals:
            signal.signal(sig, self.old_handlers[sig])

handle_signal = SignalContextProcessor


import os
import json

from sbgsdk import errors
from sbgsdk.process import process_fails
from sbgsdk.tests.common import run_cli, cleanup, RED_ADDER_ID


def test_setup():
    os.chdir(os.path.dirname(__file__))


@cleanup
@process_fails(exit_code=errors.NOT_A_JOB_ERROR)
def test_run_not_a_job():
    run_cli('run', '-i', 'input_not_a_job.json')


@cleanup
@process_fails(exit_code=errors.NO_WRAPPER_ID_ERROR)
def test_run_no_wrapper_id():
    run_cli('run', '-i', 'input_no_wrapper_id.json')


@cleanup
@process_fails(exit_code=errors.NO_RESOURCES_ERROR)
def test_run_no_resources():
    run_cli('run', '-i', 'input_no_resources.json')
    s = json.load(open('__out__.json'))
    assert s['wrapper_id'] == RED_ADDER_ID
    assert s['$$type'] == 'job'
    assert s['args']['$method'] == 'execute'


@cleanup
@process_fails(exit_code=errors.INPUTS_VALIDATION_ERROR)
def test_run_required_input_missing():
    run_cli('run', '-i', 'input_required_input_missing.json')


@cleanup
@process_fails(exit_code=errors.PARAMS_VALIDATION_ERROR)
def test_run_invalid_param():
    run_cli('run', '-i', 'input_invalid_param.json')


@cleanup
def test_run():
    run_cli('run', '-i', 'input.json')
    s = json.load(open('__out__.json'))
    assert s['wrapper_id'] == RED_ADDER_ID
    assert s['$$type'] == 'job'
    assert s['args']['$method'] == 'execute'




import uuid
from sbgsdk.exceptions import NoWrapperIdException, NoResourcesException

from sbgsdk.util import wrap_in_list
from sbgsdk.protocol import classify


class Resources(object):
    CPU_NEGLIGIBLE, CPU_SINGLE, CPU_ALL = 0, 1, -1

    def __init__(self, mem_mb=100, cpu=CPU_ALL, high_io=False):
        self.mem_mb = mem_mb
        self.cpu = cpu
        self.high_io = high_io

    def __call__(self, obj):
        obj._requirements = self
        return obj

    def __json__(self):
        return {
            '$$type': 'resources',
            'mem_mb': self.mem_mb,
            'cpu': self.cpu,
            'high_io': self.high_io,
        }

    @classmethod
    def from_dict(cls, d):
        return cls(**d)


class Outputs(object):
    def __init__(self, outputs_dict=None):
        self.outputs = outputs_dict
        for k, v in self.outputs.items():
            if isinstance(v, basestring):
                self.outputs[k] = [v]
            elif v is None:
                self.outputs[k] = []
            else:
                self.outputs[k] = list(v)

    def __json__(self):
        return {
            '$$type': 'outputs',
            'outputs': self.outputs,
        }

    @classmethod
    def from_dict(cls, d):
        return cls(outputs_dict=d.get('outputs', {}))

    def __repr__(self):
        return 'Outputs(%s)' % self.outputs


class BaseJob(object):
    PENDING, RUNNING, DONE = 'pending', 'running', 'done'

    def __init__(self, job_id=None, args=None):
        self.job_id = job_id or 'job-%s' % uuid.uuid4()
        self.args = args or {}
        self.resolved_args = {}
        self.status = BaseJob.PENDING

    def __json__(self):
        return {
            '$$type': 'job',
            'job_id': self.job_id,
            'args': self.resolved_args or self.args,
        }


class Job(BaseJob):
    def __init__(self, wrapper_id, job_id=None, resources=None, context=None, args=None,
                 method=None, inputs=None, params=None):
        if not isinstance(wrapper_id, basestring):
            wrapper_id = '.'.join([wrapper_id.__module__, wrapper_id.__name__])
        BaseJob.__init__(self, job_id, args)
        self.wrapper_id = wrapper_id
        self.context = context or {}
        self.resources = resources or Resources()
        self.args['$inputs'] = inputs or self.args.get('$inputs', {})
        self.args['$params'] = params or self.args.get('$params', {})
        self.args['$method'] = method or self.args.get('$method', None)
        self.job_id = job_id or '%s-%s-%s' % \
            (wrapper_id[wrapper_id.rfind('.') + 1:], self.args['$method'] or 'initial', uuid.uuid4())

    def __json__(self):
        return dict(BaseJob.__json__(self),
                    wrapper_id=self.wrapper_id,
                    context=self.context,
                    resources=self.resources)

    @classmethod
    def from_dict(cls, obj):
        if 'wrapper_id' not in obj:
            raise NoWrapperIdException('Wrapper ID not specified.')
        if 'resources' not in obj:
            raise NoResourcesException('Resources not specified')
        return cls(wrapper_id=obj['wrapper_id'],
                   job_id=obj.get('job_id'),
                   args=classify(obj.get('args', {})),
                   resources=classify(obj.get('resources')),
                   context=obj.get('context'))


class Join(BaseJob):
    def __init__(self, *jobs):
        BaseJob.__init__(self, args={'jobs': jobs})

    def __call__(self):
        result = {}
        # Assume dict job results are outputs struct. Join with others to act as inputs downstream.
        for job_result in filter(lambda i: isinstance(i, dict), self.resolved_args['jobs']):
            for k, v in job_result.iteritems():
                result[k] = wrap_in_list(result[k], v) if k in result else wrap_in_list(v)
        return result


class Do(BaseJob):
    def __init__(self, func, **args):
        BaseJob.__init__(self, args=args)
        self.func = func

    def __call__(self):
        return self.func(**self.resolved_args) if callable(self.func) else self.func


class Connect(BaseJob):
    def __init__(self, outputs, src_name, dst_name):
        BaseJob.__init__(self)
        self.args['outputs'] = outputs
        self.src_name = src_name
        self.dst_name = dst_name

    def __call__(self):
        return {self.dst_name: self.resolved_args['outputs'][self.src_name]}



import logging
import signal
import shlex
from pipes import quote

from sbdk.docker.container import Container
from sbdk.docker.utils import find_image, parse_repository_tag
from docker.client import APIError
from sbgsdk.util import handle_signal


class Runner(object):

    PIP_INSTALL = 'pip install {} .'
    WORKING_DIR = '/sbgenomics'

    def __init__(self, docker_client, project, **kwargs):
        self.client = docker_client
        self.project = project
        self.base_cmd = project.base_cmd
        self.sb_dir = project.sb_home
        self.working_dir = project.project_dir
        self.image_id = project.image_id
        self.dns = project.docker_dns
        self.docker_config = kwargs
        self.last_container = None

    def run(self, command, entrypoint=None, retry=5):
        volumes = {'/sb': {}, Runner.WORKING_DIR: {}}
        volumes.update(self.docker_config.pop('Volumes', {}))
        binds = {self.sb_dir: "/sb:rw",
                 self.working_dir: Runner.WORKING_DIR + ":rw"}
        binds.update(self.docker_config.pop('__binds__', {}))
        logging.info("Running command %s\nVolumes: %s\nBinds: %s", command, volumes, binds)
        config = make_config(self.image_id, command)
        config['Volumes'] = volumes
        config['WorkingDir'] = Runner.WORKING_DIR
        config['Dns'] = self.dns
        if entrypoint != None:
            config['Entrypoint'] = self.project.docker_entrypoint
        #config['PortSpecs'] = ['22']

        config.update(self.docker_config)

        try:
            cont = self.client.create_container_from_config(config)
        except APIError as e:
            # 404 means image was not found
            if e.response.status_code == 404:
                logging.info("Trying to fetch image %s" % self.image_id)
                repo, tag = parse_repository_tag(self.image_id)
                self.client.pull(repo, tag)
                image = find_image(self.client, repo, tag)
                if image is None:
                    raise
                self.image_id = image['Id']
                config['Image'] = self.image_id
                cont = self.client.create_container_from_config(config)
            else:
                raise
        try:
            self.client.start(container=cont, binds=binds)
        except APIError:
            if retry > 0:
                logging.error('Failed to run container. Retries left: %s', retry-1)
                return self.run(command, entrypoint, retry-1)
            else:
                logging.error('Failed to run container. Retries exceeded!')
                raise
        self.last_container = Container(self.client, self.project, cont)
        return self.last_container

    def run_and_print(self, command, remove=False, test=False):
        if test:
            cont = self.run(command, entrypoint=self.project.docker_entrypoint)
        else:
            cont = self.run(command)

        def handler(signum, frame):
            logging.info('Received %s: %s', signum, frame)
            cont.stop()

        with handle_signal(handler, signal.SIGTERM, signal.SIGINT):
            cont.wait()

        cont.print_log()
        if remove:
            cont.remove()

    @property
    def cmd_install(self):
        pip_options = self.project.pip_options or ''
        return Runner.PIP_INSTALL.format(pip_options)

    def chown(self, user=None, path=WORKING_DIR):
        user = user or self.project.docker_user
        return self.run(['chown', '-R', user, path])

    def run_install(self):
        return self.run(shlex.split(self.cmd_install))

    def run_wrapper(self, input_path, cwd=None, remove=True):
        run_full = self.base_cmd + ['run-full', '-i', input_path]
        if cwd:
            run_full += ['--cwd', cwd]
        self.run_and_print(multi_command(self.cmd_install, run_full), remove=remove)

    def run_job(self, input_path, output_path, cwd=None, remove=True):
        cli_run = self.base_cmd + ['run', '-i', input_path, '-o', output_path]
        if cwd:
            cli_run += ['--cwd', cwd]
        cmd = multi_command(self.cmd_install, cli_run)
        self.run_and_print(cmd, remove=remove)

    def run_tests(self, wrapper, remove=True):
        cmd = multi_command(self.cmd_install, ['nosetests', wrapper])
        self.run_and_print(cmd, remove=remove, test=True)

    def schema(self, output=None, remove=True):
        extra = ['--output', output] if output else []
        name = self.project.name
        cli_schema = self.base_cmd + ['schema'] + (['--package', name] if name else []) + extra
        cmd = multi_command(self.cmd_install, cli_schema)
        self.run_and_print(cmd, remove=remove)


class ImageRunner(Runner):

    def __init__(self, docker_client, project, **kwargs):
        Runner.__init__(self, docker_client, project, **kwargs)
        image_id = project.image_id
        info = docker_client.inspect_image(image_id)
        self.base_cmd = info['config']['Cmd']

    def run_wrapper(self, input_path, cwd=None, remove=True):
        run_full = self.base_cmd + ['run-full', '-i', input_path]
        if cwd:
            run_full += ['--cwd', cwd]
        self.run_and_print(run_full, remove=remove)

    def run_job(self, input_path, output_path, cwd=None, remove=True):
        run = self.base_cmd + ['run', '-i', input_path, '-o', output_path]
        if cwd:
            run += ['--cwd', cwd]
        self.run_and_print(run, remove=remove)

    def run_tests(self, wrapper, remove=True):
        self.run_and_print(['nosetests', wrapper], remove=remove)

    def run_install(self):
        raise NotImplementedError("Method 'run_install' not implemented for class ImageRunner.")

    def schema(self, output=None, remove=True):
        name = self.project.name if self.project else None
        cmd = self.base_cmd + ['schema'] + (['--package', name] if name else [])
        cmd += ['--output', output] if output else []
        self.run_and_print(cmd, remove=remove)


def make_config(image, command):
    return {'Image': image,
            'Cmd': command,
            'AttachStdin': False,
            'AttachStdout': False,
            'AttachStderr': False,
            'Tty': False,
            'Privileged': False,
            'Memory': 0}


def multi_command(*args):
    strings = [e if isinstance(e, basestring) else shell_join(e) for e in args]
    return ['/bin/sh', '-c', ' && '.join(strings)]


def shell_join(args):
    return ' '.join([quote(s) for s in args])






import os
import json

from sbgsdk import errors
from sbgsdk.process import process_fails
from sbgsdk.tests.common import run_cli, cleanup


def test_setup():
    os.chdir(os.path.dirname(__file__))


@cleanup
@process_fails(exit_code=errors.NO_SUCH_METHOD_ERROR)
def test_execute_no_such_method():
    run_cli('run', '-i', 'input_no_such_method.json')


@cleanup
@process_fails(exit_code=errors.NO_RESOURCES_ERROR)
def test_execute_no_resources():
    run_cli('run', '-i', 'input_execute_no_resources.json')


@cleanup
def test_execute_single():
    run_cli('run', '-i', 'input_single.json', '-o', 'output_single.json')
    s = json.load(open('output_single.json'))
    assert s['c'].endswith('/output_single.txt'), 'Unexpected output c: %s' % s['c']
    assert os.path.isfile(s['c'])




__author__ = 'sinisa'
from sbgsdk import define


class InputWrapper(define.Wrapper):
    for_each = ['inputA', 'params.group_by']

    class Outputs(define.Outputs):
        output = define.output(
            list=True,
            name="output",
            description="",
            file_types=[],
        )

    class Params(define.Params):
        path = define.string(
            name = "Input files",
            description = "Input files",
            list=True,
            )

    def execute(self):
        for param in self.params.path:
            self.outputs.output.add_file(param)._load_meta()



from sbdk.commands import check_image
from sbdk.docker import make_runner


def argument_parser(subparsers):
    parser = subparsers.add_parser("schema", help="Show schema (JSON) for all wrappers in project")

    parser.set_defaults(func=schema)


def schema(project):
        check_image(project)
        docker = make_runner(project)
        docker.schema()



from sbdk.commands import check_project_dir, check_project_data, check_project_state
from sbdk.auth import login, store_credentials
from docker.auth.auth import load_config

def argument_parser(subparsers):
    parser = subparsers.add_parser("login", help="Login to docker registry")
    parser.set_defaults(func=registry_login)


def registry_login(project):
    check_project_dir(project)
    check_project_data(project)
    check_project_state(project)

    print project.schema_server_url
    cfg = load_config()
    user, token = login(project)

    store_credentials(project.docker_registry_url, user, token, cfg)





import logging

import re
import requests
from sys import stdout

from subprocess import Popen
from docker.auth.auth import load_config

from sbdk.auth import login, store_credentials
from sbdk.error import SbdkError


ALLOWED_CHARS_RE = re.compile('^[a-z0-9]$')


def dockerize(username):
    return ''.join([chr if re.match(ALLOWED_CHARS_RE, chr) else '_' for chr in username.lower()])


def repo_name(project, user):
    saved_user = project.state.get('username')
    if not saved_user:
        saved_user = dockerize(user)
        project.state['username'] = saved_user
        project.state.save()
    else:
        saved_user = dockerize(saved_user)
    ret = project.docker_registry_name + '/' + saved_user + '/' + project.name
    return ret


class Container(object):

    def __init__(self, docker_client, project, container,
                 image=None, message=None, user=None, token=None, repository=None):
        self.docker = docker_client
        self.container = container
        self.project = project
        self.image = image
        self.message = message
        self._user = user
        self._token = token
        self._repository = repository

    @property
    def user(self):
        if not self._user:
            self.load_credentials()
        return self._user

    @property
    def token(self):
        if not self._token:
            self.load_credentials()
        return self._token

    @property
    def repository(self):
        if not self._repository:
            self._repository = repo_name(self.project, self.user)
        return self._repository

    def inspect(self):
        return self.docker.inspect_container(self.container)

    def is_running(self):
        return self.inspect()['State']['Running']

    def wait(self):
        if self.is_running():
            self.docker.wait(self.container)
        return self

    def is_success(self):
        self.wait()
        return self.inspect()['State']['ExitCode'] == 0

    def remove(self):
        self.wait()
        self.docker.remove_container(self.container)

    def stop(self, nice=False):
        return self.docker.stop(self.container) if nice else self.docker.kill(self.container)

    def print_log(self):
        if self.is_running():
            for out in self.docker.attach(container=self.container, stream=True):
                print(out.rstrip())
        else:
            print(self.docker.logs(self.container))

    def commit(self, message=None, config=None, tag=True):
        if message:
            self.message = message

        self.image = self.docker.commit(self.container['Id'], message=message, conf=config)
        if tag and self.project.name and self.user:
            self.docker.tag(self.image['Id'], self.repository, tag=self.tag)
        else:
            print("Image ID:", self.tag)
        return self

    def save_as_base(self):
        if not self.image:
            raise SbdkError("Container image not committed yet!")
        self.project.state['image_id'] = self.image['Id']
        self.project.state.save()

    def push(self):
        if not self.image:
            raise SbdkError("Container image not committed yet!")
        push = Popen(['docker', 'push', self.repository], stdout=stdout)
        push.wait()

    def load_credentials(self, renew=False):
        cfg = load_config()
        session_id = cfg.get(self.project.docker_registry_url, None).get('password', None) if cfg.get(
            self.project.docker_registry_url, None) else None
        resp = requests.get(self.project.schema_server_url + '/auth', headers={"session-id": session_id}, verify=False)

        if renew or resp.status_code != 200:
            self._user, self._token = login(self.project)
            store_credentials(self.project.docker_registry_url, self._user, self._token, cfg)

        else:
            self._user = cfg[self.project.docker_registry_url]['username']
            self._token = cfg[self.project.docker_registry_url]['password']


    def check_privilege(self):
        if not self.project.state.get('username'):
            raise SbdkError("File state.json doesn't contain username")
        url = '%s/repo/%s/%s' %(self.project.schema_server_url, dockerize(self.project.state['username']),
                                self.project.state['project_name'])
        resp = requests.get(url, headers={"session-id": self._token}, verify=False)
        contributors = resp.json()['contributors'] if resp.json().get('contributors', None) else []
        if (resp.status_code == 404 and self.project.state.get('username') == self._user) or (
                        resp.status_code == 200 and (dockerize(self._user) == resp.json()['owner'] or self._user in contributors)):
            return True
        return False


    @property
    def tag(self):
        return self.image['Id'][:16]


from os.path import join, exists
import json


class State(dict):
    """ Utility class to handle sbdk state file. NOT THREAD SAFE """

    FILENAME = 'state.json'

    def __init__(self, path):
        self.state_file = join(path, State.FILENAME)
        d = {}
        if exists(self.state_file):
            with open(self.state_file) as f:
                d = json.load(f)
        dict.__init__(self, d)

    def save(self):
        with open(self.state_file, 'w') as f:
            json.dump(self, f)

    def __enter__(self):
        return self

    def __exit__(self):
        self.save()



from uuid import uuid4
import os
from sbgsdk import define, require
from sbgsdk.tests.test_tools.adder import adder


@require(200, cpu=require.CPU_NEGLIGIBLE)
class RedAdder(define.Wrapper):
    class Inputs(define.Inputs):
        a = define.input(required=True, list=True)
        b = define.input(required=True, list=True)

    class Params(define.Params):
        n = define.integer(default=1, min=1)

    class Outputs(define.Outputs):
        output = define.output(list=True)

    def _out_file_name(self, sample):
        return 'sum_of_'+sample+'.txt'

    def execute(self):
        assert self.get_allocated_memory('MB') == 200, 'expected resources mismatch (%s, 200)' % self.resources.mem_mb
        if self.params.n > 1:
            return self.execute_parallel()

        by_sample_a = {}
        for i in xrange(len(self.inputs.a)):
            key = self.inputs.a[i].meta['sample']
            by_sample_a[key] = self.inputs.a[i]

        by_sample_b = {}
        for i in xrange(len(self.inputs.b)):
            key = self.inputs.b[i].meta['sample']
            by_sample_b[key] = self.inputs.b[i]

        for k in by_sample_a.keys():
            adder.add(by_sample_a[k], by_sample_b[k], self._out_file_name(k))
            out = self.outputs.output.add_file(self._out_file_name(k))
            out.meta = {'sample': k}

    def execute_parallel(self):
        return self.job('split', requirements=require(300, cpu=require.CPU_SINGLE))

    def split(self):
        assert self.get_allocated_memory('MB') == 300, 'expected resources mismatch (%s, 300)' % self.resources.mem_mb
        a_files = map(os.path.abspath, adder.split(self.inputs.a[0], self.params.n))
        b_files = map(os.path.abspath, adder.split(self.inputs.b[0], self.params.n))
        return self.job('unsplit',
                        {'files': [self.job('add', {'files': [a, b]}) for a, b in zip(a_files, b_files)]},
                        requirements=require(mem_mb=400, cpu=require.CPU_ALL))

    @require(500)
    def add(self, files):
        assert self.get_allocated_memory('MB') == 500, 'expected resources mismatch (%s, 500)' % self.resources.mem_mb
        temp_out = str(uuid4()) + ".txt"
        adder.add(files[0], files[1], temp_out)
        return os.path.abspath(temp_out)

    def unsplit(self, files):
        adder.unsplit(self._out_file_name(), files)
        self.outputs.output.add_file(self._out_file_name())


import os
import json
import logging
import tempfile
import shutil
from copy import deepcopy

from sbgsdk.util import DotAccessDict


class IOValue(unicode):
    meta = None

    def __new__(cls, value=None):
        if isinstance(value, list):
            value = value[0] if value else ''
        value = os.path.abspath(value) if value else value
        obj = unicode.__new__(cls) if not value else unicode.__new__(cls, value)
        obj.meta = DotAccessDict()
        return obj

    file = property(lambda self: self)

    def _load_meta(self):
        metadata_file = self + '.meta'
        if not os.path.exists(metadata_file):
            logging.warning('No metadata file found: %s', metadata_file)
            self.meta = DotAccessDict()
            return
        logging.debug('Loading metadata from %s', metadata_file)
        with open(metadata_file) as f:
            content = json.load(f)
        if not isinstance(content, dict):
            logging.error('Metadata not a dict: %s', content)
            self.meta = DotAccessDict()
        else:
            self.meta = DotAccessDict(**content)

    def _save_meta(self):
        if not self.file:
            return
        metadata_file = self.file + '.meta'
        logging.debug('Saving metadata to %s', metadata_file)
        if not isinstance(self.meta, dict):
            logging.error('Metadata not a dict: %s', self.meta)
            self.meta = {}
        if not os.path.exists(metadata_file):
            with open(metadata_file, 'w') as f:
                json.dump(self.meta or {}, f)
            return
        tmp = tempfile.mktemp()
        with open(tmp, 'w') as f:
            json.dump(self.meta or {}, f)
        shutil.move(tmp, metadata_file)

    def _abspath(self):
        self.file = os.path.normpath(os.path.abspath(self.file))

    @property
    def size(self):
        return os.stat(self).st_size if os.path.exists(self) else 0

    def make_metadata(self, **kwargs):
        result = deepcopy(self.meta)
        result.update(**kwargs)
        return result




import os
import nose.tools
from sbgsdk.tests.test_tools.generator import generator
from sbgsdk.tests.test_tools.comparator import comparator


def remove_txt_files():
    for name in os.listdir('.'):
        if name.endswith('.txt'):
            os.remove(name)


@nose.tools.with_setup(remove_txt_files, remove_txt_files)
@nose.tools.raises(Exception)
def test_compare_random():
    count = 1000
    generator.generate('random_a.txt', count, 'random')
    generator.generate('random_b.txt', count, 'random')
    comparator.compare('random_a.txt', 'random_b.txt', 'random_c.txt')


@nose.tools.with_setup(remove_txt_files, remove_txt_files)
def test_compare_same():
    count = 1000
    generator.generate('same_a.txt', count, 'same', 45)
    generator.generate('same_b.txt', count, 'same', 45)
    comparator.compare('same_a.txt', 'same_b.txt', 'same_c.txt')

    f = open('same_c.txt')
    first_line = f.readline()
    count2 = int(first_line.rstrip())
    nose.tools.assert_equals(count, count2)

    line_count = 0
    for line in f:
        integer_num = int(line.rstrip(), 16)
        assert(integer_num == 45)
        line_count += 1
    nose.tools.assert_equals(line_count, count)

    f.close()


@nose.tools.with_setup(remove_txt_files, remove_txt_files)
@nose.tools.raises(Exception)
def test_compare_same_diff_length():
    generator.generate('same_a.txt', 1000, 'same', 45)
    generator.generate('same_b.txt', 1001, 'same', 45)
    comparator.compare('same_a.txt', 'same_b.txt', 'same_c.txt')



from sbdk.commands import check_image, check_project_dir
from sbdk.docker import make_runner


def argument_parser(subparsers):
    parser = subparsers.add_parser("test", help="Test wrapper")
    parser.set_defaults(func=test)
    parser.add_argument("module", help="Python module with tests. E.g. 'sbg_wrappers.my_wrapper'. "
                                       "Same as nosetests '--where'", type=str, default='.')
    parser.add_argument("--keep-container", "-k", help="Don't remove container when run finishes", type=bool)


def test(project, module, keep_container=False):
    check_project_dir(project)
    check_image(project)
    r = make_runner(project)
    try:
        r.run_tests(module, remove=not keep_container)
    finally:
        c = r.chown()
        c.remove()


__test__ = False




import os
import json
import unittest
from sbgsdk.tests.common import run_cli, remove_temp_files


class Test(unittest.TestCase):
    def setUp(self):
        os.chdir(os.path.dirname(__file__))

    def tearDown(self):
        remove_temp_files()

    def test_schema(self):
        p = run_cli('schema')
        s = json.loads(p.stdout.get_string())
        self.assertIsInstance(s, list)
        if s:
            self.assertIn('wrapper_id', s[0])

    def test_run_full(self):
        run_cli('run-full', '-i', 'input_full.json')
        s = json.load(open('__out__.json'))
        self.assertTrue(s['c'].endswith('output_parallel.txt'))
        self.assertTrue(os.path.isfile(s['c']))

    def test_execute_error(self):
        run_cli('run', '-i', 'input_error.json', '-o', 'output_error.json')
        s = json.load(open('output_error.json'))
        self.assertEqual(s['$$type'], 'error')
        self.assertIn('No such file', s['message'])


if __name__ == '__main__':
    unittest.main()


from .schema_based import SchemaBased
from .base_attr import BaseAttr
from .validators import validate_file_type, validate_file_exists
from .io_value import IOValue
from .io_list import IOList


class IOAttr(BaseAttr):
    def __init__(self, name=None, description='', file_types=None, required=False, list=False, **extra):
        BaseAttr.__init__(self, name, description, required, list=list, **extra)
        self.types = file_types or []

    def __call__(self, attr_name, data=None):
        data = BaseAttr.__call__(self, attr_name, data)
        return IOList(data) if self.list else (IOValue(data) if data else None)

    def validate(self, instance):
        BaseAttr.validate(self, instance)
        if instance:
            validate_file_type(instance, self.types, list=self.list)
            validate_file_exists(instance, list=self.list)
        elif self.required:
            raise ValueError('This input is required')


class IODef(SchemaBased):
    def __setattr__(self, key, value):
        attr_def = getattr(self.__class__, key)
        if not attr_def:
            raise AttributeError('%s does not define %s' % (self.__class__.__name__, key))
        if value is None:
            return object.__setattr__(self, key, None)
        if attr_def.list and isinstance(value, list):
            return SchemaBased.__setattr__(self, key, IOList(value))
        if not attr_def.list and not isinstance(value, IOValue):
            return SchemaBased.__setattr__(self, key, IOValue(value))
        if isinstance(value, (IOValue, IOList)):
            return SchemaBased.__setattr__(self, key, value)
        raise ValueError('Unsupported type: %s' % type(value))

    def _save_meta(self):
        for v in self:
            if v.value:
                v.value._save_meta()

    def _load_meta(self):
        for v in self:
            val = getattr(self, v.id)
            if val and hasattr(val, '_load_meta'):
                val._load_meta()



import os
import sys
import logging
from subprocess import Popen, PIPE

from sbdk.auth import login


def argument_parser(subparsers):
    parser = subparsers.add_parser("upload", help="Upload a file to sbgenomics.")
    parser.set_defaults(func=upload)

    parser.add_argument("file", type=str, help="Local path of file to upload")


def upload(project, file):
    executable = project.vs_client_path
    server = project.file_server_url
    if not executable or not server:
        print('Uploader location or file server URL missing.')
        sys.exit(1)
    if not os.path.isfile(file):
        print('%s is not a file.' % file)
        sys.exit(1)
    file = os.path.abspath(file)
    usr, sid = login(project)
    user_home = '/Stash/%s' % usr
    dest = os.path.join(user_home, os.path.basename(file))
    cwd = os.path.abspath('.')
    args = [executable, '-o', 'u', '-sessionId', '-s', server, '-ssl', '-lf', file, '-rf', dest]
    logging.debug('Running cmd: %s', args)
    os.chdir('/tmp')
    try:
        p = Popen(args, stdin=PIPE)
        p.communicate('%s\n' % sid)
    finally:
        os.chdir(cwd)


import re
import os


def validator(func):
    def f(val, *args, **kwargs):
        if not kwargs.pop('list'):
            return func(val, *args, **kwargs)
        for v in val:
            func(v, *args, **kwargs)
    return f


# noinspection PyUnusedLocal
@validator
def validate_type(val, t_tuple, **kw):
    if val is None:
        return
    if not isinstance(val, t_tuple):
        raise ValueError('Expected %s, got %s' % (' or '.join([t.__name__ for t in t_tuple]), type(val).__name__))


# noinspection PyUnusedLocal
@validator
def validate_range(val, min=None, max=None, step=None, **kw):
    if val is None:
        return
    if min is not None and val < min:
        raise ValueError('Value cannot be less than %s. Got %s.' % (min, val))
    if max is not None and val > max:
        raise ValueError('Value cannot be more than %s. Got %s.' % (max, val))


# noinspection PyUnusedLocal
@validator
def validate_pattern(val, pattern, **kw):
    if val is None or pattern is None:
        return
    if not re.match(pattern, val):
        raise ValueError('Value "%s" does not match pattern "%s."' % (val, pattern))


# noinspection PyUnusedLocal
@validator
def validate_enum(val, valid_vals, **kw):
    if val is None:
        return
    if val not in valid_vals:
        raise ValueError('Invalid value. Expected %s. Got %s.' % (' or '.join(valid_vals), val))


# noinspection PyUnusedLocal
@validator
def validate_file_type(val, types, **kw):
    if val is None:
        return
    file_type, types = val.meta.file_type, types
    if types and file_type not in types:
        raise ValueError('File type is %s. Expected %s' % (file_type, ' or '.join(types)))


# noinspection PyUnusedLocal
@validator
def validate_file_exists(val, **kw):
    if val is None:
        return
    if not os.path.exists(val.file):
        raise ValueError('File does not exist: %s' % val.file)


# noinspection PyUnusedLocal
@validator
def validate_struct(val, **kw):
    if val is None:
        return
    errors = val._validate()
    if errors:
        raise ValueError(str(errors))


import os.path
from pkgutil import get_data


def read(name):
    return get_data('sbdk.resources', name)


def copy(name,  dst):
    joined = os.path.join(dst, os.path.basename(name))
    dst_name = joined if os.path.isdir(dst) else dst
    if os.path.exists(dst_name):
        print "File '{}' already exists, skipping".format(dst_name)
        return
    with open(dst_name, 'w') as out:
        out.write(read(name))


def process_template(file_name, dst=None, **bind_data):
    tpl = read(file_name)
    out = tpl.format(**bind_data)
    if dst:
        dst_name = dst
        if os.path.exists(dst) and os.path.isdir(dst):
            dst_name = os.path.join(dst, file_name)
        if os.path.exists(dst_name):
            print "File '{}' already exists, skipping".format(dst_name)
            return out
        with open(dst_name, 'w') as of:
            of.write(out)
    return out



import os
import json

from sbgsdk.tests.common import run_cli, cleanup, RED_ADDER_ID


def test_setup():
    os.chdir(os.path.dirname(__file__))

@cleanup
def test_execute_parallel():
    run_cli('run', '-i', 'input_parallel.json', '-o', 'output_parallel.json')
    s = json.load(open('output_parallel.json'))
    assert s['wrapper_id'] == RED_ADDER_ID
    assert s['$$type'] == 'job'
    assert s['args']['$method'] == 'split'


@cleanup
def test_execute_split():
    run_cli('run', '-i', 'input_split.json', '-o', 'output_split.json')
    s = json.load(open('output_split.json'))
    assert s['$$type'] == 'job'
    assert s['args']['$method'] == 'unsplit'
    assert len(s['args']['files']) == 10


@cleanup
def test_execute_add():
    run_cli('run', '-i', 'input_add.json', '-o', 'output_add.json')
    s = json.load(open('output_add.json'))
    assert isinstance(s, basestring) and s.endswith('.txt'), 'Unexpected output: %s' % s


@cleanup
def test_execute_unsplit():
    run_cli('run', '-i', 'input_unsplit.json', '-o', 'output_unsplit.json')
    s = json.load(open('output_unsplit.json'))
    assert s['c'].endswith('/output_parallel.txt'), 'Unexpected output c: %s' % s['c']
    assert os.path.isfile(s['c'])




import pkg_resources
import argparse
import os
import sys
import logging
import logging.config

from sbgsdk.errors import INTERNAL_ERROR
from sbgsdk.exceptions import exit_code_for_exception, NotAJobError, ValidationException, JobException, ProtocolException
from sbgsdk.protocol import from_json, to_json
from sbgsdk.job import Job
from sbgsdk.util import import_name
from sbgsdk.wrapper import Wrapper
from sbgsdk.executor import Tassadar


def create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--logging-config', help='Path to json file for logging.config.dictConfig', default='')
    subparsers = parser.add_subparsers()

    sch = subparsers.add_parser('schema', help='Print schema for all wrappers on stdout')
    sch.set_defaults(cmd_func=cmd_schema)
    sch.add_argument('--package', help='Look for wrappers in this package.')
    sch.add_argument('--output', help='Print schema to this file.', default='')

    run = subparsers.add_parser('run', help='Run wrapper job.')
    run.set_defaults(cmd_func=cmd_run, run_method=run_job)
    run.add_argument('--cwd', default='.', help='cd here before running job.')
    run.add_argument('-i', '--input', default='__in__.json', help='JSON file that contains arguments for wrapper job.')
    run.add_argument('-o', '--output', default='__out__.json', help='Where to write the job result.')

    run = subparsers.add_parser('run-full', help='Run whole wrapper.')
    run.set_defaults(cmd_func=cmd_run, run_method=run_wrapper)
    run.add_argument('--cwd', default='.', help='cd here before running job.')
    run.add_argument('-i', '--input', default='__in__.json', help='JSON file with entry point args')
    run.add_argument('-o', '--output', default='__out__.json', help='Where to write the outputs.')

    return parser


def get_wrapper_schemas(package=None):
    group = 'sbgsdk.wrappers'
    map_id_class = {}
    classname = lambda wrp_cls: '.'.join([wrp_cls.__module__, wrp_cls.__name__])
    if package:
        pkg = import_name(package)
        for var in dir(pkg):
            obj = getattr(pkg, var)
            if isinstance(obj, type) and issubclass(obj, Wrapper):
                map_id_class[classname(obj)] = obj

    for entry_point in pkg_resources.iter_entry_points(group=group):
        wrp_cls = entry_point.load()
        full_class_name = classname(wrp_cls)
        map_id_class[full_class_name] = wrp_cls
    return [dict(schema=v._get_schema(), wrapper_id=k) for k, v in map_id_class.iteritems()]


def run_job(job):
    job.resolved_args = job.args
    return Tassadar().exec_wrapper_job(job)


def run_wrapper(job):
    return Tassadar().execute(job)


# noinspection PyUnusedLocal
def cmd_schema(output, **kwargs):
    sch = get_wrapper_schemas(kwargs.pop('package', None))
    if output:
        with open(output, 'w') as fp:
            to_json(sch, fp)
    else:
        print(to_json(sch))


# noinspection PyUnusedLocal
def cmd_run(cwd, input, output, run_method=run_job, **kwargs):
    if not os.path.isdir(cwd):
        raise Exception('No such directory: %s', cwd)
    os.chdir(cwd)

    if not os.path.isfile(input):
        raise Exception('No such file: %s' % input)
    with open(input) as fp:
        job = from_json(fp)
    if not isinstance(job, Job):
        raise NotAJobError('Input JSON must describe a job.')

    try:
        result = run_method(job)
    except JobException as e:
        result = e

    with open(output, 'w') as fp:
        to_json(result, fp)


def run_script():
    args = vars(create_parser().parse_args())
    if args['logging_config']:
        with open(args['logging_config']) as fp:
            logging.config.dictConfig(from_json(fp))
    else:
        logging.getLogger().setLevel(logging.DEBUG)
    try:
        args['cmd_func'](**args)
    except ProtocolException as e:
        logging.exception("Bad Job: %s", args)
        return exit_code_for_exception(e)
    except ValidationException as e:
        logging.exception("Input or params validation failed: %s", args)
        return exit_code_for_exception(e)
    except JobException as e:
        logging.exception("Wrapper error: %s", args)
        return exit_code_for_exception(e)
    except Exception:
        logging.exception("Internal error: %s", args)
        return INTERNAL_ERROR
    return 0


if __name__ == '__main__':
    sys.exit(run_script())



import os
import nose.tools
from sbgsdk.tests.test_wrappers.green_generator import GeneratorWrapper
from sbgsdk.tests.common import cleanup


def remove_temp_files():
    for name in os.listdir('.'):
        if name.endswith('.txt') or name.endswith('.meta'):
            os.remove(name)


cleanup = nose.tools.with_setup(remove_temp_files, remove_temp_files)


@cleanup
def _check_file(filename, count, condition):
    f = open(filename)
    first_line = f.readline()
    count2 = int(first_line.rstrip())
    nose.tools.assert_equals(count, count2)

    line_count = 0
    for line in f:
        integer_num = int(line.rstrip(), 16)
        assert(condition(integer_num))
        line_count += 1
    nose.tools.assert_equals(line_count, count)

    f.close()


@cleanup
def test_generator_random():
    w = GeneratorWrapper(
        inputs={},
        params={
            'number_of_files': 1,
            'numbers_per_file': 1000,
            'template_number': 3,
            'numbers_to_generate': 'random'
        }
    )

    w.test()

    filename = w.outputs.output[0].file
    count = w.params['numbers_per_file']
    check = lambda x: 0 <= x < 256
    _check_file(filename, count, check)


@cleanup
def test_generator_same():
    w = GeneratorWrapper(
        inputs={},
        params={
            'number_of_files': 1,
            'numbers_per_file': 1000,
            'template_number': 44,
            'numbers_to_generate': 'same'
        }
    )

    count = w.params['numbers_per_file']

    w.test()

    filename = w.outputs.output[0].file
    count = w.params['numbers_per_file']
    check = lambda x: x == 44
    _check_file(filename, count, check)


@cleanup
def test_generator_two_files():
    w = GeneratorWrapper(
        inputs={},
        params={
            'number_of_files': 2,
            'numbers_per_file': 1000,
            'template_number': 3,
            'numbers_to_generate': 'random'
        }
    )

    count = w.params['numbers_per_file']

    w.test()

    filename0 = w.outputs.output[0].file
    count0 = w.params['numbers_per_file']
    check0 = lambda x: 0 <= x < 256
    _check_file(filename0, count0, check0)

    filename1 = w.outputs.output[1].file
    count1 = w.params['numbers_per_file']
    check1 = lambda x: 0 <= x < 256
    _check_file(filename1, count1, check1)



from setuptools import setup, find_packages
import sys
import os.path

sys.path.append(os.path.dirname(__file__))

from sbdk import VERSION


setup(
    name="sbgsdk",
    version=VERSION,
    include_package_data=True,
    packages=find_packages(),
    entry_points={
        'console_scripts': ['sbg = sbdk.main:main'],
    },
    install_requires=['nose==1.3.0', 'docker-py==0.3.1']
)


import filecmp
from sbgsdk.job import Job
from sbgsdk.protocol import from_json, to_json
from sbgsdk.executor import Tassadar
from sbgsdk.tests.test_tools.generator import generator
from sbgsdk.tests.test_wrappers import RedAdder
from sbgsdk.tests.common import cleanup


@cleanup
def test_red_adder():
    inputs = {'a': ['a.txt'], 'b': ['b.txt']}
    generator.generate('a.txt')
    generator.generate('b.txt')

    job_single = Job(RedAdder, inputs=inputs, params={'n': 1})
    Tassadar().execute(job_single)

    job_multi = Job(RedAdder, inputs=inputs, params={'n': 5})
    Tassadar().execute(from_json(to_json(job_multi)))

    assert filecmp.cmp('output_parallel.txt', 'output_single.txt')


import itertools


class AttrDef(object):
    _count = itertools.count(0) # Total number of attributes defined

    def __init__(self):
        self._order = AttrDef._count.next()

    def __call__(self, attr_name, data=None):
        """ Override to instantiate an attribute. """
        pass

    def validate(self, instance):
        """ Override to validate an attribute. """
        pass

    def get_schema(self):
        def get_value(val):
            return val._get_schema() if callable(getattr(val, '_get_schema', None)) else val
        return dict([(k,get_value(v)) for k,v in self.__dict__.iteritems() if k[0] != '_'])


import functools
import itertools
import operator

from sbgsdk.wrapper import Wrapper
from sbgsdk import require


class SimpleParallelWrapper(Wrapper):
    for_each = '', ''

    def _entry_point(self):
        return self.job('split')

    @require(100, require.CPU_NEGLIGIBLE)
    def split(self):
        input_id, metadata_key = self.for_each
        inp = getattr(self.inputs, input_id)

        if metadata_key.startswith('params.'):
            metadata_key = getattr(self.params, metadata_key[len('params.'):])
        if metadata_key not in ('sample_group', 'sample', 'library', 'platform_unit', 'chunk', 'file', None):
            raise Exception('Invalid metadata_key')

        groups = group_inputs(inp, metadata_key)
        reqs = self.get_requirements()
        jobs = [self.job('work', requirements=reqs, args={'group_id': k, 'files': files})
                for k, files in groups.iteritems()]
        return self.job('merge', args={'job_results': jobs})

    def work(self, group_id, files):
        # Remove files from input_id that are not in job
        inp = getattr(self.inputs, self.for_each[0])
        inp._values = filter(lambda f: f.file in files, inp._values)

        # Call wrapper
        self.execute()
        self.outputs._save_meta()

        # Serialize outputs to job result
        result = self.outputs.__json__()

        for key, val in result.items():
            result[key] = val.__json__()
        return result

    @require(100, require.CPU_NEGLIGIBLE)
    def merge(self, job_results):
        for result in job_results:
            for output_id, files in result.iteritems():
                for file_path in files:
                    getattr(self.outputs, output_id).add_file(file_path)._load_meta()


S = '__!__'


def make_rg_id(metadata_key, io_obj):
    rg = [
        io_obj.meta.sample_group or '',
        io_obj.meta.sample or '',
        io_obj.meta.library or '',
        io_obj.meta.platform_unit or '',
        str(io_obj.meta.chunk) if io_obj.meta.chunk is not None else '',
    ]
    rg_map = {
        'sample': rg[:2],
        'library': rg[:3],
        'platform_unit': rg[:4],
        'chunk': rg[:5],
    }
    return S.join(rg_map[metadata_key]) if metadata_key in rg_map else getattr(io_obj.meta, metadata_key)


def group_inputs(inp, metadata_key):
    if str(metadata_key) == 'None':
        return {'': [f for f in inp]}
    if metadata_key == 'file':
        return {f: [f] for f in inp}
    key_getter = functools.partial(make_rg_id, metadata_key)
    files = sorted(inp, key=key_getter)
    return {key: [f.file for f in val] for key, val in itertools.groupby(files, key_getter)}


def match(io_obj, job_id):
    if has_blank_meta(io_obj):
        return False
    job_id = job_id or ''
    chunk_id = make_rg_id('chunk', io_obj)
    return zipmatch(chunk_id.split(S), job_id.split(S))


def rtrim_iterable(iterable_):
    """
    >>> rtrim_iterable([1, 2, 0, 0])
    [1, 2]
    """
    return list(reversed(list(itertools.dropwhile(operator.not_, reversed(iterable_)))))


def zipmatch(a, b):
    """
    >>> zipmatch([1, 2, 0], [1, 2])
    True
    >>> zipmatch([1, 2, 3], [1, 2])
    False
    """
    return all(x[0] == x[1] for x in zip(rtrim_iterable(a), rtrim_iterable(b)))


def has_blank_meta(io_obj):
    return make_rg_id('chunk', io_obj) == S.join(['', '', '', '', ''])



from .attrdef import AttrDef


def default_attr_name(name):
    if not isinstance(name, basestring):
        name = str(name)
    if name.startswith('-') or name.isupper():
        return name
    return name.replace('_', ' ').capitalize()


class BaseAttr(AttrDef):
    def __init__(self, name=None, description='', required=False, list=False, **extra):
        AttrDef.__init__(self)
        self.name = name
        self.description = description
        self.required = required
        self.list = list
        self._extra = extra

    @property
    def extra(self):
        return self._extra

    def __call__(self, attr_name, data=None):
        if attr_name.startswith('_'):
            raise ValueError('Input, output and parameter IDs cannot start with "_" (%s)' % attr_name)
        if not self.name:
            self.name = default_attr_name(attr_name)
        if not self.description:
            self.description = self.name
        return data

    def validate(self, instance):
        from io_list import IOList
        if self.list and not isinstance(instance, (tuple, list, IOList)):
            raise ValueError('Value must be tuple or list.')
        if not self.required:
            return
        if instance is None:
            raise ValueError('You must specify a value.')
        if self.list and not instance:
            raise ValueError('You must specify a value.')
